# Wobbly Balls
# Copyright (C) 2010 Jeffrey Armstrong <jeff@rainbow-100.com>
# http://jeff.rainbow-100.com/
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# A full version of the license terms is available in LICENSE.

import pygame
import random
import math
from constants import *

def scaled_color(fraction):
    col = None
    if fraction <= 0.5:
        c = int(fraction*2.0*255.0)
        col = pygame.Color(255,max(0,min(255,c)),0,255)
    elif fraction <= 1.0:
        c = int(255.0*(1.0-2.0*(fraction-0.5)))
        col = pygame.Color(max(0,min(255,c)),255,0,255)
    else:
        col = pygame.Color('white')
    return col
        
def strength_color(strength):
    fraction = float(strength-S_MIN)/float(S_MAX-S_MIN)
    return scaled_color(fraction)

def weight_color(weight):
    fraction = float(weight-W_MIN)/float(W_MAX-W_MIN)
    return scaled_color(fraction)

class Column:
    def __init__(self):
        self.blocks = []
        
    def add_block(self,block):
        self.blocks.append(block)
        
    def broken_blocks(self):
        broken = []
        for i in range(0,len(self.blocks)):
            weight = 0
            for j in range(i+1,len(self.blocks)):
                weight = weight + self.blocks[j].weight
            if self.blocks[i].is_broken(weight):
                broken.append(self.blocks[i])
        return broken
        
    def remove(self,block):
        self.blocks.remove(block)
        
    def full(self):
       return len(self.blocks) >= MAX_COL_HEIGHT
       
    def empty(self):
        self.blocks = []

class GColumn(Column):
    def __init__(self, centerx, bottomy, rocker):
        Column.__init__(self)
        
        self.centerx = centerx
        self.bottomy = bottomy
        self.x = centerx
        self.y = bottomy
        self.xoffset = None
        self.yoffset = None
        
        self.rocker = rocker
        
        self.angle = 0
        self.aup = True
    
    def set_angle(self,angle):
        self.angle = -1.0*angle
        self.aup = True
        
    def add_block(self,block):
        self.aup = True
        self.blocks.append(block)

    def empty(self):
        for b in self.blocks:
            self.rocker.block_rendering.remove(b)
        self.blocks = []

    def get_top(self):
        
        if self.xoffset == None:
            self.update()
        
        #xbase = self.x - self.rocker.rect.centerx - self.rocker.roll_dist
        xbase = self.x
        #xoffset = xbase*math.cos(math.radians(self.angle))
        #yoffset = xbase*math.sin(math.radians(self.angle))
                        
        xoffset = self.xoffset + (COL_HEIGHT+10)*math.sin(math.radians(self.angle))
        yoffset = self.yoffset - (COL_HEIGHT+10)*math.cos(math.radians(self.angle))
            
        x = self.rocker.rect.centerx + xoffset # - self.rocker.roll_dist
        y = self.rocker.rect.centery + yoffset
        
        
        
        #x = self.xoffset + (COL_HEIGHT+10)*math.sin(math.radians(self.angle))
        #y = self.yoffset + self.bottomy - (COL_HEIGHT+10)*math.cos(math.radians(self.angle))
        #y = self.yoffset - (COL_HEIGHT+10)*math.cos(math.radians(self.angle))
                
        #x = x + self.rocker.rect.centerx #-self.rocker.roll_dist
        #y = self.rocker.rect.centery + y 
        
        return (x,y)

    def update(self):
        
        xbase = self.x #- self.rocker.rect.centerx
        
        if self.aup:
            self.xoffset = xbase*math.cos(math.radians(self.angle))
            self.yoffset = xbase*math.sin(math.radians(self.angle))
            
            # Iterate through blocks
            dx = 2
            for x in self.blocks:
                x.rect.centerx = self.xoffset + dx*math.sin(math.radians(self.angle))
                x.rect.bottom = self.yoffset - dx*math.cos(math.radians(self.angle))
                
                x.rect.centerx = x.rect.centerx + self.rocker.rect.centerx
                x.rect.bottom = x.rect.bottom + self.rocker.rect.centery 
                
                # Increment dx
                dx = dx + BLOCK_Y + 2

class Block(pygame.sprite.Sprite):
    def __init__(self,weight=None,strength=None):
        pygame.sprite.Sprite.__init__(self)

        if strength:
            self.strength = strength
        else:
            self.strength = random.randint(S_MIN,S_MAX)
            
        if weight:
            self.weight = weight
        else:
            self.weight = random.randint(W_MIN,W_MAX)
        
    def is_broken(self,weight):
        return (weight > self.strength)
    
class GBlock(Block):
    
    def __init__(self,weight=None,strength=None):
        Block.__init__(self, weight, strength)
    
        self.image = pygame.Surface((BLOCK_X, BLOCK_Y))
        self.rect = pygame.Rect(0,0,BLOCK_X, BLOCK_Y)
        
        # Draw the "block"
        self.image.convert()
        self.image.set_colorkey(pygame.Color('black'))
        self.image.fill(pygame.Color('black'))
        
        pygame.draw.ellipse(self.image,weight_color(self.weight),self.rect,0)
        pygame.draw.ellipse(self.image,strength_color(self.strength),self.rect,3)
        
class PlayBlockGroup:
    def __init__(self,level,rocker):
        
        self.blocks = []
        for i in range(0,level.blocks):
            self.blocks.append(GBlock(weight=level.weight(),strength=level.strength()))
            
        self.cposition = 0
        
        self.rocker = rocker
        
    def right(self):
        if self.cposition + len(self.blocks) < NUM_COLUMNS:
            self.cposition = self.cposition + 1
    
    def left(self):
        if self.cposition > 0:
            self.cposition = self.cposition - 1

    def update(self):
        for ic in range(self.cposition, self.cposition + len(self.blocks)):
            self.blocks[ic-self.cposition].rect.midbottom = self.rocker.columns[ic].get_top()
                        
    def allowed(self):
        allow = True
        for ic in range(self.cposition, self.cposition + len(self.blocks)):
            if self.rocker.columns[ic].full():
                allow = False
                break
                
        return allow        
    
    def drop(self):
        for ic in range(self.cposition, self.cposition + len(self.blocks)):
            self.rocker.columns[ic].add_block(self.blocks[ic-self.cposition])
            
        self.blocks = None
        
    def random_drop(self):
        """If called, the current play_group is attempted to be placed.  If
        not possible after three attempts, a column will be destroyed."""
        
        for x in self.blocks:
            tries = 0
            full_c = None
            while tries < 3:
                ic = random.randint(0,NUM_COLUMNS-1)
                if not self.rocker.columns[ic].full():
                    self.rocker.columns[ic].add_block(x)
                    break
                elif full_c == None:
                    full_c = ic
                tries = tries + 1
            
            if tries >= 3 and  full_c != None:
                self.rocker.columns[full_c].empty()
            
        self.blocks = None
